/***********************************************************
* Project:	CS298                                          *
* File:		GDIToolbox1.cpp                                *
* Purpose:	Partial implementation of Simple Raster        *
*			Graphics Package library for Windows is defined*
*			in this file.  Functions available in this file*
*			help users to set the colors, or linestyle of  *
*			objects, or create lines, ellipses, rectangles,*
*			or texts.  This file also contain information  *
*			about canvases that are offscreen bitmaps or   *
*			pixmaps of the requisite size.                 *
************************************************************/

#include "gditoolbox1.hpp"
#include <math.h> //for cos and sin in EllipseArc

/* Partial implementation of Simple Raster Graphics Package (GDI) library,
for Windows. Version 1.1.  Copyright (C) 1996 Rudy Rucker.
See header for log information.*/

//---Constructor and Destructor---------------------------------
GDIParams::GDIParams():
	_textcolor(START_TEXTCOLOR),
	_linecolor(START_LINECOLOR),
	_linewidth(START_LINEWIDTH),
	_markerradius(START_MARKERRADIUS),
	_linestyle(START_LINESTYLE),
	_fillcolor(START_FILLCOLOR),
	_fillstyle(START_FILLSTYLE),
	_hatchstyle(START_HATCHSTYLE),
	_blankcolor(START_BLANKCOLOR),
	_drawmode(START_DRAWMODE)
{}

GDIToolbox::GDIToolbox():_hdc(NULL)
{
/* The default GDIParams constructor initializes _params.  The only
other work we need to do is to make the _hpen, _hbrysh, and _hblankbrush*/
	_hpen = CreatePen(_params._linestyle, _params._linewidth, _params._linecolor);
	LOGBRUSH logbrush = {_params._fillstyle, _params._fillcolor,
		_params._hatchstyle};
	_hbrush = CreateBrushIndirect(&logbrush);
	_hblankbrush = CreateSolidBrush(_params._blankcolor);
	_hrgn = NULL; //Means no clipping
}

GDIToolbox::~GDIToolbox()
{ /* Assume that this happens AFTER the _hdc has been invalidated.
	So you don'd need to select the tools out of the _hdc.*/
   DeleteObject(_hpen);
   DeleteObject(_hbrush);
   DeleteObject(_hblankbrush);
   if (_hrgn)
	DeleteObject(_hrgn);
}
//------------Accessor---------
void GDIToolbox::GetParams(GDIParams &par)
{
	par = _params; /* The default overloaded operator= for the
   	GDIParam class will copy the structure field by field. */
}

//------------Mutators----------
void GDIToolbox::SetHDC(HDC hdc)
{ //ALWAYS need to do this to get a valid _hdc before making other GDI calls.
	_hdc = hdc;
   if (!_hdc)
   	return;
	SelectObject(_hdc, _hpen);
	SelectObject(_hdc, _hbrush);
	SelectClipRgn(_hdc, _hrgn);
	SetBkMode(_hdc, TRANSPARENT);  //For text. TRANSPARENT defn in windows.h
	::SetTextColor(_hdc, _params._textcolor); //Windows function, not GDIToolbox
	SetROP2(_hdc, _params._drawmode);
}

void GDIToolbox::SetParams(GDIParams &par)
{
	SetTextColor(par._textcolor);
	SetLineColor(par._linecolor);
	SetFillColor(par._fillcolor);
	SetBlankColor(par._blankcolor);
	SetFillStyle(par._fillstyle, par._hatchstyle);
	SetLineWidth(par._linewidth);
	SetLineStyle(par._linestyle);
	SetDrawMode(par._drawmode);
 }

void GDIToolbox::SetTextColor(COLORREF color)
{
	if (_params._textcolor == color)
		return;
	_params._textcolor = color;
	if (_hdc)
		::SetTextColor(_hdc, _params._textcolor);
}

void GDIToolbox::SetLineColor(COLORREF color)
{
	HPEN holdpen = _hpen;

	if (_params._linecolor == color)
		return;
	_params._linecolor = color;
	_hpen = CreatePen(_params._linestyle, _params._linewidth, _params._linecolor);
	if (_hdc)
		SelectObject(_hdc, _hpen);
	DeleteObject(holdpen);
}

void GDIToolbox::SetFillColor(COLORREF color)//Sets the fill or background
{
	HBRUSH holdbrush = _hbrush;
	LOGBRUSH logbrush;

	if (_params._fillcolor == color)
		return;
	logbrush.lbStyle = _params._fillstyle;
	logbrush.lbColor = _params._fillcolor = color;
	logbrush.lbHatch = _params._hatchstyle;
	_hbrush = CreateBrushIndirect(&logbrush);
   if (_hdc)
	   SelectObject(_hdc, _hbrush);
	DeleteObject(holdbrush);
}

void GDIToolbox::SetBlankColor(COLORREF color)//Sets the blank background
{
	if (_params._blankcolor == color)
		return;

	DeleteObject(_hblankbrush);
	_params._blankcolor = color;
	_hblankbrush = CreateSolidBrush(_params._blankcolor);
}

void GDIToolbox::SetLineStyle(int style) //Some predefined lineStyles are
{
	HPEN holdpen = _hpen;

	if (_params._linestyle == style)
		return;
	_params._linestyle = style;
	_hpen = CreatePen(_params._linestyle, _params._linewidth, _params._linecolor);
	if (_hdc)
	 	SelectObject(_hdc, _hpen);
	DeleteObject(holdpen);
}

void GDIToolbox::SetLineWidth(int width) //width is in pixels, use like 1 to 10.
{
	HPEN holdpen = _hpen;
	if (_params._linewidth == width)
		return;
	_params._linewidth = width;
	_hpen = CreatePen(_params._linestyle, _params._linewidth, _params._linecolor);
   if (_hdc)
		SelectObject(_hdc, _hpen);
	DeleteObject(holdpen);
}

void GDIToolbox::SetFillStyle(int fillstyle, int hatchstyle)
{
	LOGBRUSH logbrush;
	HBRUSH holdbrush = _hbrush;

	if (fillstyle == _params._fillstyle &&  hatchstyle == _params._hatchstyle)
		return;
	logbrush.lbStyle = _params._fillstyle = fillstyle;
	logbrush.lbColor = _params._fillcolor;
	logbrush.lbHatch = _params._hatchstyle = hatchstyle;

	_hbrush = CreateBrushIndirect(&logbrush);
	if (_hdc)
		SelectObject(_hdc, _hbrush);
	DeleteObject(holdbrush);
}

void GDIToolbox::SetDrawMode(int drawmode)
{
	_params._drawmode = drawmode;
	if (_hdc)
	   SetROP2(_hdc, _params._drawmode);
}

//----Graphics output methods ---------------------------------
void GDIToolbox::SetPixel(int x, int y, COLORREF color)
{
	if (!_hdc)
	   	return;
	::SetPixel(_hdc,x,y,color); //Use the Windows SetPixel, not GDIToolBox::SetPixel
}

void GDIToolbox::SetPixel(Point target, COLORREF color)
{
	SetPixel(target.x(), target.y(), color);
}

void GDIToolbox::Line(int x1, int y1, int x2, int y2)
{   //Draws a line in current line color, width, style.
	if (!_hdc)
   		return;
	MoveToEx(_hdc, x1, y1, NULL); //Use for Win32 compatability
	LineTo(_hdc, x2, y2);
	return;
}

void GDIToolbox::Line(Point p1, Point p2)
{
	Line(p1.x(), p1.y(), p2.x(), p2.y());
}

void GDIToolbox::Polyline(int count, Point * varray)
{
	if (!_hdc)
   		return;
	switch (count)
	{
		case 0:
			return;
		case 1:
			Line(varray[0], varray[0]); //GDIToolbox::Line function.
			break;
		default:
			::Polyline(_hdc, (POINT *)varray, count);
	/*The :: scope here means use the global Windows Polyline function, not
		the GDToolbox::Polyline function. */
			break;
	}
}

void GDIToolbox::Polygon(int count, Point * varray)
{
	if (!_hdc)
   		return;
	::Polygon(_hdc, (POINT *)varray, count);
	/*The :: scope here means use the global Windows Polygon function, not
		the GDIToolbox::Polygon function. */

}

void GDIToolbox::Rectangle(int left, int bottom, int right, int top)
{
	if (!_hdc)
   		return;
	::Rectangle(_hdc, left, top, right, bottom); //Calls Windows Rectangle funct.
}

void GDIToolbox::Rectangle(Point lefttop, Point rightbottom)
{
	Rectangle(lefttop.x(), lefttop.y(), rightbottom.x(), 
		rightbottom.y()); //GDIToolbox::Rectangle
}

void GDIToolbox::Rectangle(Rect rect)
{
	Rectangle(rect.left(), rect.bottom(), rect.right(), rect.top());
		//GDIToolbox::Rectangle
}

void GDIToolbox::Ellipse(int left, int top, int right, int bottom)
{
	if (!_hdc)
	  	return;
	::Ellipse(_hdc, left, top, right, bottom);
}

void GDIToolbox::Ellipse(Point lefttop, Point rightbottom)
{
	Ellipse(lefttop.x(), lefttop.y(), rightbottom.x(), 
		rightbottom.y()); //GDIToolbox::Ellipse
}

void GDIToolbox::Ellipse(Rect rect)
{
	Ellipse(rect.left(), rect.bottom(), rect.right(), rect.top());
		//GDIToolbox::Rectangle
}

void GDIToolbox::EllipseArc(Rect rect, float startangle, float endangle)
{
	if (!_hdc)
   		return;
	Arc(_hdc, rect.left(), rect.bottom(),
		rect.right(), rect.top(),
		int(((rect.left() + rect.right())/2.0) + 1000.0 * cos(startangle)),
		int(((rect.bottom() + rect.top())/2.0) + 1000.0 * sin(startangle)),
		int(((rect.left() + rect.right())/2.0) + 1000.0 * cos(endangle)),
		int(((rect.bottom() + rect.top())/2.0) + 1000.0 * sin(endangle)));
	/* The last four arguments of Arc are int coords of points that lie on
	the line of the starting angle and the ending angle.  We multiply by 1000.0
	just so as not to lose too much info to round off.*/
}

void GDIToolbox::Circle(Point center, Real radius)
{
	Ellipse(center.x() - (int)radius, center.y() - (int)radius,
		center.x() + (int)radius, center.y() + (int)radius); //GDIToolbox::Ellipse
}

void GDIToolbox::Marker(Point center)
{
	Circle(center, _params._markerradius);
}

void GDIToolbox::Text(Point origin, char* text)
{
	if (!_hdc)
		return;
	TextOut(_hdc, origin.x(), origin.y(), text, lstrlen(text));
}

void GDIToolbox::Clear(int left, int top, int right, int bottom)
{
	if (!_hdc)
   		return;
	SelectObject(_hdc, _hblankbrush);
	PatBlt(_hdc, left, top, right - left, bottom - top, PATCOPY);
		/*PatBlt covers the rect with the selected brush pattern. Args:
     	target, target corner, rect width and height, raster operation.*/
 	SelectObject(_hdc, _hbrush);  //Back to normal.
    return;
}

void GDIToolbox::Clear(Rect rect)
{
	Clear(rect.left(), rect.top(), rect.right(), rect.bottom());
}

void GDIToolbox::FloodFill(Point startpixel)
{
	if (!_hdc)
		return;
	ExtFloodFill(_hdc, startpixel.x(), startpixel.y(),
		GetPixel(_hdc, startpixel.x(), startpixel.y()), FLOODFILLSURFACE);
}
#define DISK_SPRAY

void GDIToolbox::Spray(Point center, Real density, Real radius)
{
	int count;
	Real x, y;

#ifdef DISK_SPRAY
	count = (int)(density * PI * radius * radius); //area of circle
#else //square
	count = (int)(density * 4.0 * radius * radius); //area of square
#endif //DISK_SPRAY
	if (!count) //If density is low, you could get count 0 and nothing would appear.
		count = 1;
	if (!_hdc)
		return;
	for (int i=0; i<count; i++)
	{	/* Get a random point (x,y) with vector magnitude betwen 0 and 1, then
		add radius*(x,y) to center.*/
#ifdef DISK_SPRAY //Spray a disk.
		Randomunitdiskpair(&x, &y);
		::SetPixel(_hdc, center.x() + (int)(radius * x),
			center.y() + (int)(radius * y), _params._fillcolor);
#else //This is a faster method, it sprasys a square.
		//Randomunitdiskpair(&x, &y);
		::SetPixel(_hdc, center.x() - (int)radius + Random((int)(2*radius)),
			center.y() - (int)radius + Random((int)(2*radius)),
			_params._fillcolor);
#endif
	} 
}

//Clip methods---------------------------

void GDIToolbox::SetClipRectangle(Rect cliprect)
{
	HRGN holdrgn = _hrgn;

	_hrgn = CreateRectRgn(cliprect.left(),cliprect.top(),cliprect.right(),
		cliprect.bottom());
	if (_hdc)
		SelectClipRgn(_hdc, _hrgn);
	DeleteObject(holdrgn);
}

void GDIToolbox::SetClipRectangle(HWND hwnd)
{
	RECT clientrect;

   	GetClientRect(hwnd, &clientrect);
	SetClipRectangle(Rect(clientrect));
}

Rect GDIToolbox::GetClipRectangle()
{
	RECT cliprect;

	if (!_hrgn) //NULL means no clipping.  So return a screen-sized rect.
		return Rect(0, 0, GetSystemMetrics(SM_CXFULLSCREEN), 
			GetSystemMetrics(SM_CYFULLSCREEN));
	else
	{
		GetRgnBox(_hrgn, &cliprect);
		return Rect(cliprect);
	}
}


