/***********************************************************
* Project:	CS298                                          *
* File:		GDIToolbox1.hpp                                *
* Purpose:	Partial implementation of Simple Raster        *
*			Graphics Package library for Windows is defined*
*			in this file.  Functions available in this file*
*			help users to set the colors, or linestyle of  *
*			objects, or create lines, ellipses, rectangles,*
*			or texts.  This file also contain information  *
*			about canvases that are offscreen bitmaps or   *
*			pixmaps of the requisite size.                 *
************************************************************/

#ifndef GDIWIN_HPP
#define GDIWIN_HPP
#include "types.h" //for <windows.h> and for Real
#include "point2.hpp"
#include "rect1.hpp"
#include "random.h" //for Spray

/*-----Instructions for use-----
Create a global variable of the Toolbox.  This can also be
a static inside WndProc, if no other procedure writes graphics.
This declaration calls the constructor.

GDIToolbox gdibox;

You can use the GDIToolbox mutators to Set the colors, linestyle, fillflag
at any time. These mutators do not care if _hdc is currently valid ---
although you should make sure to SetHDC the _hdc to NULL whenever it
becomes invalid (i.e. by a Release(hwnd, hdc) or an EndPaint(hwnd)).
Note that the default start value of _hdc is indeed NULL.
You might want to call some mutators in WM_CREATE to change the
initialization.

The graphics methods will only write things after you have called
gdibox.SetHDC(hdc) on an hdc which is valid.

If you are going to use the toolbox object gdibox for an direct
screen write, do this for each group of writes:

	hdc = GetDC(hwnd);
	gdibox.SetHDC(hdc);
	//various gdibox.??? screen writes
	ReleaseDC(hwnd, hdc);
	gdibox.SetHDC(NULL);

The reason for the gdibox.SetHDC(NULL) is so that the gdibox
can know that it doesn't have a valid _hdc anymore.  As mentioned
above, you can still use the mutators on the gdibox object even when
its _hdc is NULL.

If you are going to use a WindowBitmap *wbm object, then do this
in WM_CREATE:

	wbm = new WindowBitmap(hwnd);
	gdibox.SetHDC(wbm->hdc());
   //possibly use GDI mutators to Set the colors, linestyle, fillflag.

When you are through using the WindowBitmap (possibly not until
WM_DESTROY) make these calls:

	delete wbm;
	gdibox.SetHDC(NULL);

When the program ends, gdibox goes out of scope and its desctructor is
called, releasing its tools _hpen, _hbrush, and _hblankbrush.
*/

//GDI graphic output functions  ---------------------------------
#define START_TEXTCOLOR  RGB(0,0,0)
#define START_LINECOLOR  RGB(255,255,255)
#define START_LINEWIDTH   1
#define START_MARKERRADIUS   5
#define START_LINESTYLE   PS_SOLID
#define START_FILLCOLOR  RGB(0,255,0)
#define START_FILLSTYLE  BS_SOLID
#define START_HATCHSTYLE 0
#define START_BLANKCOLOR RGB(255,255,255)
#define START_DRAWMODE R2_COPYPEN

class GDIParams
{
private:
	COLORREF _textcolor;
	COLORREF _linecolor;
	int _linewidth;
	int _markerradius;
	int _linestyle;
	COLORREF _fillcolor;
	int _fillstyle;
	int _hatchstyle;
	COLORREF _blankcolor;
	int _drawmode;
	friend class GDIToolbox; //So this class can use the private fields.
public:
	GDIParams();
};

class GDIToolbox
{
private:
	GDIParams _params;
	HDC _hdc; //_hdc is the active device context
	HPEN _hpen;
	HBRUSH _hbrush, _hblankbrush;
	HRGN _hrgn; //clipping rect region.
public:
//---Constructors and Destructor  ---------------------------------
	GDIToolbox();
	~GDIToolbox();
//---Accessors ---------------------
	void GetParams(GDIParams &par);
  	COLORREF GetTextColor(){return _params._textcolor;}
	COLORREF GetLineColor(){return _params._linecolor;}
	COLORREF GetFillColor(){return _params._fillcolor;}
	COLORREF GetBlankColor(){return _params._blankcolor;}
	int GetFillStyle(){return _params._fillstyle;}
	int GetHatchStyle(){return _params._hatchstyle;}
	int GetLineWidth(){return _params._linewidth;}
	int GetMarkerRadius(){return _params._markerradius;}
	int GetPenStyle(){return _params._linestyle;}
	int GetDrawMode(){return _params._drawmode;}
//----Mutators ---------------------------------
	void SetHDC(HDC hdc);
	void SetParams(GDIParams &par);
	void SetLineColor(COLORREF color); /* This Sets the _linecolor.
		Standard GDI calls this SetColor.
		You Set COLORREF values by using the Windows RGB(r,g,b) macro, where
		r,g, and b are BYTEs coding the relative intensitied of red, green,
		and blue in the final color.  The BYTEs range from 0 (the least)
		to 255 (the most). */
	void SetTextColor(COLORREF color); //This Sets the _text color.
	void SetFillColor(COLORREF color);/*Sets the _fillcolor.  Standard
		GDI calls this function SetBackGroundColor */
	void SetBlankColor(COLORREF color);/* Sets the blankcolor used for
		clearing the display */
	void SetLineStyle(int style); /* Use the Windows lineStyle codes:
		PS_SOLID, PS_DASH, PS_DOT, PS_DASHDOT, PS_DASHDOTDO, PS_NULL.  More
		info in Windows documenation on CreatePen.*/
	void SetLineWidth(int width); //width is in pixels, use like 1 to 10.
	void SetMarkerRadius(int markerradius){_params._markerradius = markerradius;} /*Pixel
		radius of a disk used for a marker */
	void SetFillStyle(int fillstyle, int hatchstyle = 0); /* The first argument
		is a windows can be BS_SOLID, BS_HOLLOW, BS_HATCHED for a fill that is respectively
		solid, non-existent, or hatched.  The second argument can be a hatchstyle
		HS_HORIZONTAL, HS_VERTICAL, HS_BDIAGONAL (hatch lines have positive slope),
		HS_FDIAGONAL (lines have negative slope), HS_CROSS, HS_DIAGCROSS.  More
		info in Windows documentation on CreateBrushIndirect.*/
	void SetDrawMode(int drawmode); /* the drawmodes are the
		raster-operation codes defined in the SetROP2 API documnetation.
	   R2_XORPEN is xor write, and R2_COPYPEN is normal write.*/
//----Graphics methods--------------------------------------
	/*Standard GDI assumes the window coordinates start at (0,0) at lower left
	 corner,but to make life easier, we'll instead use the  Windows and DOS
	assumption that the window coordiantes start at (0,0) at the upper left
	corner, with the y count increasing as you move DOWN the screen.  All
	the graphic functions except pixel use the _linecolor to
	draw their lines, and uses the _fillcolor to fill interiors.*/
	void SetPixel(int x, int y, COLORREF color); /* Sets a single pixel.*/
	void SetPixel(Point target, COLORREF color); /* Sets a single pixel.*/
	void Line(int x1, int y1, int x2, int y2);
	void Line(Point p1, Point p2);
	void Polyline(int count, Point* varray);
	void Polygon(int count, Point* varray);
		/*Draws a polyline plus a line from varray[0] to varray[count-1],
		and fills the polygon */
	void Rectangle(int left, int bottom, int right, int top);//filled rect
	void Rectangle(Point lefttop, Point rightbottom);//filled rect
	void Rectangle(Rect rect);//filled rect.
	void Circle(Point center, Real radius); //Radius is a pixel measure.
	void Marker(Point center); //Circle of radius _marker_radius
	void Ellipse(int left, int bottom, int right, int top);//Filled ellipse in rect.
	void Ellipse(Point lefttop, Point rightbottom); //The rect is the bounding box.
	void Ellipse(Rect rect); 
	void EllipseArc(Rect rect, float startangle, float endangle); /*Angles
		between 0 and 2*PI are the start and end radius angle of the arc swept
		out.  The rect is the bounding box of the ellipse.*/
	void Text(Point origin, char* text); /*  Uses _textcolor.  To use
		the function, you must make sure the text array is	null-terminated.
	   (A declared constant string like "this" is null-terminated.
	   The origin is the left bottom. */
	void Clear(int left, int top, int right, int bottom); //Fill rect with blankcolor
	void Clear(Rect rect); //Fills rect with blankcolor
	void FloodFill(Point start); /*Puts fillcolor on all pixels which connect
		to startpixel in an unbroken chain of pixels with the color of startpixel*/
	void Spray(Point center, Real density, Real radius); /* Density should be
		between 0.0 and 1.0.  Spray fills a density proportion of the pixels
		in the radius disk around center with fillcolor */
	void SetClipRectangle(Rect cliprect);
	void SetClipRectangle(HWND hwnd);
	Rect GetClipRectangle();
};
#endif //GDIWIN_HPP
