/***********************************************************
* Project:	CS298                                          *
* File:		camera3.cpp                                    *
* Purpose:	The camera, face culling, projection from 3D to*
*			2D are defined in this file.                   *
************************************************************/
#include "camera3.hpp"
#include <math.h>


void Camera::Initialize(const Box3 &worldbox)
{
	_from = Vector3(0.0, 0.0, 1.0);
	_up = Vector3(0.0, 1.0, 0.0);
	_eye = worldbox.Center() + Vector3(0.0, 0.0, 2.0 * worldbox.ZRadius());
		// Put the eye out past the front xy face of the box.
	_viewrect = Box2(worldbox.lox(), worldbox.loy(), worldbox.hix(),
		worldbox.hiy());
		//Put in some viewrect the shape of the worldbox xz face.
	SetProjectionType(PERSPECTIVE_NORMAL); /*This sets _projectiontype
		then calls SetViewAngle which sets _viewangle and calls
		Set Params which sets _eyedist and _viewcenter and then calls
		calls SetMatrices to set _rotation and _view.*/
}

void Camera::SetEye(Vector3 neweye)
{
	_eye = neweye;
     _SetParams();
}

void Camera::SetFromUp(Vector3 newfrom, Vector3 newup)
{
	_from = newfrom;
	_up = newup;
	_SetParams();
}

void Camera::SetEyeFromUp(Vector3 neweye,Vector3 newfrom, Vector3 newup)
{
	_eye = neweye;
	_from = newfrom;
	_up = newup;
	_SetParams();
}

void Camera::SetViewAngle(Real angle)
{
	if (angle < MIN_VIEWANGLE)
		angle = MIN_VIEWANGLE; //Don't want a tan of 0.
	if (angle > MAX_VIEWANGLE)
		angle = MAX_VIEWANGLE; //Don't want an infinite tan.
	_viewangle = angle;
	_SetParams();
}

void Camera::_SetParams()
{
	Real viewradius = MIN(_viewrect.XRadius(), _viewrect.YRadius());
	_eyedist = viewradius / tan(_viewangle / 2.0); /*Explanation:
	/*Think of a right triangle with _eye at the "x" corner, the
	_vieworigin at the right angle, and the closest edge of the
	_viewrect at the "y" corner.  This adjusts the _eyedist so as
	to make _viewrect just the right size relative to the camera's
	x and y.  You do need to adjust the z-location of the projection
	plane that contains the _viewrect, and you do this now:*/
	_viewcenter = _eye - _eyedist * _from;
	_SetMatrices();
}

void Camera::_SetMatrices()
{ /* Use the _from, _up, and _viewcenter values to set _rotation and _view*/
	_rotation.Identity();
	_rotation.SetColumns((_up*_from).Normalize(),_up,_from);
	_view.SetTranslationRotation(_viewcenter, _rotation);
	_view.Invert();
}

void Camera::SetProjectionType(int projectiontype)
{
	switch(projectiontype)
	{
		case PARALLEL:
			_projectiontype = PARALLEL;
			break;
		case PERSPECTIVE_TELEPHOTO:
			_projectiontype = PERSPECTIVE_TELEPHOTO;
			_viewangle = TELEPHOTO_VIEWANGLE;
			break;
		case PERSPECTIVE_WIDEANGLE:
			_projectiontype = PERSPECTIVE_WIDEANGLE;
			_viewangle = WIDEANGLE_VIEWANGLE;
			break;
		case PERSPECTIVE_NORMAL:
		default:
			_projectiontype = PERSPECTIVE_NORMAL;
			_viewangle = NORMAL_VIEWANGLE;
			break;
	}
	_SetParams();
}

void Camera::MatchViewRectAspect(int cx, int cy)
{
	_viewrect.MatchAspect(cx, cy);
	_SetParams();
}

void Camera::XBump(int updown)
{
	_eye += updown*DIST_STEP*(_up * _from); //The camera's own x-axis.
     _SetParams();
}

void Camera::YBump(int updown)
{
	_eye += updown*DIST_STEP*_up;//the camera's own y-axis.
     _SetParams();
}

void Camera::ZBump(int updown)
{
	_eye += updown*DIST_STEP*_from; //the camera's own z-axis.
     _SetParams();
}

/*For all of these rotations, you can think of it this way.  You have
a vector V (_from or _up) that you want to rotate into a rotated_V by
some rotation matrix extra_rot. To rotate V in some extra_rot way, you
can rotate V back to its start position V_start by looking at
V_start = _rotation_inverse * V, then rotate V_start by extra_rot, and then
multiply that result by _rotation to restore the initial rotation.
In other words,
the rotated_V = _rotation * extra_rot * _rotation_inverse * V.
But from is just _rotation * k, with k the unit Vector3(0.0, 0.0, 1.0).
Substituting, we get
rotated_from = _rotation * extra_rot * _rotation_inverse * _rotation * k.
Cancelling, you get
rotated_from = _rotation * extra_rot * k.
The same reasoning applies to _up.
*/

void Camera::XBumpRotate(int updown)
{
	_from = _rotation * XRotationMatrix(updown*ANGLE_STEP) *
		Vector3(0.0, 0.0, 1.0);  //Tip the k vector, then move it to here
	_up = _rotation * XRotationMatrix(updown*ANGLE_STEP) *
		Vector3(0.0, 1.0, 0.0);  //Tip the j vector, then move it to here
     _SetParams();
}

void Camera::YBumpRotate(int updown)
{
	_from = _rotation * YRotationMatrix(updown*ANGLE_STEP) *
		Vector3(0.0, 0.0, 1.0);  //Tip the k vector, then move it to here
	_up = _rotation * YRotationMatrix(updown*ANGLE_STEP) *
		Vector3(0.0, 1.0, 0.0);  //Tip the j vector, then move it to here
     _SetParams();
}

void Camera::ZBumpRotate(int updown)
{
	_from = _rotation * ZRotationMatrix(updown*ANGLE_STEP) *
		Vector3(0.0, 0.0, 1.0);  //Tip the k vector, then move it to here
	_up = _rotation * ZRotationMatrix(updown*ANGLE_STEP) *
		Vector3(0.0, 1.0, 0.0);  //Tip the j vector, then move it to here
     _SetParams();
}

BOOL Camera::SphereClipTest(const Vector3 &testcentroid, Real testradius)
{
/*The strategy here is imagine a sphere of vertices located around
testcentroid at a radius of testradius.  We return FALSE if (a) any
of the vertices might be closer than TOO_CLOSE_TO_EYE, or if (b) ALL of
the vertices will project to points outside of _viewrect.  If (a) doesn't
hold but only SOME of the projected vertices will lie in _viewrect, we
return a TRUE anyway because we can let the hardware do the clipping
against _viewrect*.  In other words, we return TRUE if the sphere of
vertices isn't too close and if some of them are visible.*/
	Real projectfactor;
	Vector2 projectcentroid;

	if (_projectiontype == PARALLEL)
	{
		projectcentroid.Set(testcentroid.x(), testcentroid.y());
		if (_viewrect.Distance(projectcentroid) > testradius)
		// Then no part of sphere can be in the _viewrect.   See comment below.
			return FALSE;
		else
			return TRUE;
	}
	else //_projectiontype == PERSPECTIVE
	{
		//First set projectfactor to distance between _eye and _centroid
		projectfactor = _eyedist - testcentroid.z();
		if (projectfactor < TOO_CLOSE_TO_EYE + testradius)
		/*This means that some part of sphere may be closer than
		TOO_CLOSE_TO EYE.  The idea is that at the worst, a vertex might
		be directly testradius units closer than testcentroid.*/
			return FALSE;
		projectfactor = _eyedist / projectfactor;
		projectcentroid.Set( projectfactor * testcentroid.x(),
			projectfactor * testcentroid.y());
		if (_viewrect.Distance(projectcentroid) > projectfactor * testradius)
		/* Then no part of the sphere can be in the _viewrect.  This is
		because the x and y directions of p are both scaled in the same
		way.  So the furthest a projected vertex can be from projcentroid
		is projectfactor*testradius.  If the _viewrect.Distance is greater
		than this, then no vertex of the sphere can be inside _viewrect.*/
			return FALSE;
		else
			return TRUE;
	}
}

void Camera::Project(Poly3 &p)
{
	Real projectfactor;

	if (!p.Visible()) //Then you failed the ClipTest.
		return;
	if (_projectiontype == PARALLEL)
		for (int i=0; i<p.Count(); i++)
		{
			p._projectpoly._vector2vert[i].Set(
				p._vector3vert[i].x(), p._vector3vert[i].y());
		}
	else //_projectiontype == PERSPECTIVE
		for (int i=0; i<p.Count(); i++)
		{
			//Convert to the coordinate of the projection plane system.
			projectfactor = _eyedist/ (_eyedist - p._vector3vert[i].z());
			p._projectpoly._vector2vert[i].Set(
				projectfactor * p._vector3vert[i].x(),
				projectfactor * p._vector3vert[i].y());
		}
}

void Camera::Project(Polyhedron3 &p)
{ //Be careful not to do this until after p passes SphereClipTest
	Real projectfactor;

	if (_projectiontype == PARALLEL)
		for (int i=0; i<p.Count(); i++)
		{
			p._vector2vert[i].Set(
				p._vector3vert[i].x(), p._vector3vert[i].y());
		}
	else //_projectiontype == PERSPECTIVE
		for (int i=0; i<p.Count(); i++)
		{
			//Convert to the coordinate of the projection plane system.
			projectfactor = _eyedist/ (_eyedist - p._vector3vert[i].z());
			p._vector2vert[i].Set(
				projectfactor * p._vector3vert[i].x(),
				projectfactor * p._vector3vert[i].y());
		}
}

void Camera::BackFaceCull(Polyhedron3 &p)
{
	Vector3 cameraeye(0.0, 0.0, _eyedist);
	//The coordinates of the _eye in the camera system
	Vector3 face_to_eye;
	for (int i=0; i<p._facecount; i++)
	{
		if (_projectiontype == PARALLEL)
			face_to_eye = Vector3(0.0, 0.0, 1.0);
		else // PERSPECTIVE
			face_to_eye = (cameraeye - p._facecentroid[i]).Normalize();
/*		p._facedistance[i] = -(face_to_eye.ReturnMagnitudeAndNormalize());
Put the - so that smaller value means farther away.  Later I might
want to change Poly3::SegmentCopy(Polyhedron p, int faceindex) so it
puts p._facedistance[i] into the _centroid._z field for a perhaps more
accurate depth sort.  But then I'd have to do the same thing for plain
Poly3 objects in space, and currently this is not being done.*/
		if (face_to_eye % p._facenormal[i] >= 0.0)
			p._facevisible[i] = TRUE;
		else
			p._facevisible[i] = FALSE;
	}
}

void Camera::ZClipTest(Polyhedron3 &p)
{
/* The idea here is to keep only the polygons whose verts
are all in front of your eye.  We use the _facedistance field
to signal the good (1.0) and bad (-1.0) faces.  And we turn
the bad vertices into dumb zero verts just so that its save
to freely Project and RealToPixel after doing a ZClipTest*/

	Real hither = _eyedist - TOO_CLOSE_TO_EYE;
	BOOL *goodvertindex = new BOOL[p._count];

     for (int i=0; i< p._count; i++)
	{
		if (p._vector3vert[i].z() > hither)
		{
			goodvertindex[i] = FALSE;
			p._vector3vert[i].Set(0.0, 0.0, 0.0); //So it's
				//safe to Project and RealToPixel this vert.
		}
		else
			goodvertindex[i] = TRUE;
	}
	for (i=0; i< p._facecount; i++)
	{
		p._facedistance[i] = 1.0; //assume in front of me.
		for (int j=0; j<p._facevertcount[i]; j++)
			if (!(goodvertindex[p._face[i][j]]))
			{
				p._facedistance[i] = -1.0; //means behind me.
				break; //out of j loop
			}
	}
	delete [] goodvertindex;
}

void Camera::SetRealWindow(Frame &frame)
{
	frame.SetRealWindow(_viewrect.lox(),_viewrect.loy(),_viewrect.hix(),
		_viewrect.hiy());
}
 


//-------------Fourth dimension stuffs-------
void Camera::InternalFaceCull(Polyhedron3 &p)
{
	p.SetVertInHull();
	p.SetInternalFace();

	for (int i=0; i<p._facecount; i++)
	{
		if (p._internalface[i]==FALSE)
			p._facevisible[i] = TRUE;
		else
			p._facevisible[i] = FALSE;
	}

}

