/***********************************************************
* Project:	CS298                                          *
* File:		camera3.hpp                                    *
* Purpose:	The camera, face culling, projection from 3D to*
*			2D are defined in this file.                   *
************************************************************/

#ifndef CAMERA_HPP
#define CAMERA_HPP
#include "types.h"
#include "polyhedron4.hpp"
#include "box2.hpp"
#include "frame1.hpp"

/*	We think of a camera as a view-plane and a center of projection
called COP or _eye.
	The scene is projected onto the view-plane as follows:  given a point
P in the scene we draw a line from _eye to P and let P' be the point where
the line crosses the view-plane.
	The view-plane is specified by (1) a _viewcenter point called
view reference point or VRP and by (2) the _from vector which is the normal
to the view-plane also known as the VPN.
	If you want to center some target point T in your view, then _from
should be (_eye - T).Normalize().
	The y-axis of the view-plane is the _up vector, also known as the
view-up vector or VUP.  Ordinarly in the world we think of the z-axis as
up, so we default _up to match the world's z-direction.
	The _eye = _view_plane_origin + _eyedist*_from, that is,
_eye is _eyedist units away from _view_plane_origin along the normal to the
view-plane.  Conversely, _view_plane_origin = _eye - _eyedist*+_from; you
use this when you change _from, but leave _eye fixed.
	The visible part of the view-plane is the _viewrect.  _viewrect is
set to match the screen window.  We adjust _eyedist so that if viewradius is
the smaller of the half-axis of viewrect, then viewradius/_dist =
tan(_viewangle).
	The _rotation matrix tracks the camera's rotational aspect, that is,
it's column vectors are _from*_up, _up, and _from.  The _view matrix is the
inverse of the rigid-body motion aspect matrix gotten by perfoming the
_rotation transfomration followed by a translation to _viewcenter.

*/

//Flags for _projectiontype
#define PARALLEL 0
#define PERSPECTIVE_TELEPHOTO 1
#define PERSPECTIVE_NORMAL 2
#define PERSPECTIVE_WIDEANGLE 3
//Ranges to clamp
#define MIN_VIEWANGLE (PI/20.0)
#define MAX_VIEWANGLE (PI - (PI/4.0))
//Defaults
#define TOO_CLOSE_TO_EYE 0.01  //Hither clip distance
#define TOO_FAR_FROM_EYE 3.0  //Yon clip distance
#define ANGLE_STEP (PI/20.0) //For the BumpRotate
#define DIST_STEP 1.0
#define TELEPHOTO_VIEWANGLE (PI/6.0)  //Changed 4/30/96
#define NORMAL_VIEWANGLE (PI/3.0)
#define WIDEANGLE_VIEWANGLE (2.0*PI/3.0)

class Camera
{
private:
	Vector3 _from, _up, _viewcenter, _eye;//Foley&VanDam: VPN, VUP, VRP, COP.
	Box2 _viewrect; //_viewrect sets _viewradius, which sets _eyedist.
	Real _viewangle, _eyedist; //_viewangle and _viewrect set _eyedist*/
	Matrix4 _rotation; //The xyz rotation aspect of the camera.
	Matrix4 _view; //Converts to camera reference system coordinates.
	int _projectiontype; /*PARALLEL, PERSPECTIVE_TELEPHOTO,
		PERSPECTIVE_NORMAL,	or PERSPECTIVE_WIDEANGLE*/
	void _SetParams(); //helper functions
	void _SetMatrices(); /*Uses _from and _up to set _rotation, then uses
		_rotation and _viewcenter to set _view*/
public:
	//Constructors
	Camera(){}//Default constructor
	Camera(const Box3 &worldbox){Initialize(worldbox);} //Other constructor
	void Initialize(const Box3 &worldbox); //Helper
	//Mutators
	void MatchViewRectAspect(int cx, int cy);
	void SetViewAngle(Real angle);/*sets _viewangle, uses _viewrect to set
		_eyedist, then uses _eye and _eyedist to set
		_viewcenter, then calls SetMatrices to adjust _view*/
	void BumpViewAngle(int updown); //updown should be +1 or -1.
	void XBump(int updown); //These move _eye and adjust everything else.
	void YBump(int updown);
	void ZBump(int updown);
	void XBumpRotate(int updown); //These move _from & _up, fix every else.
	void YBumpRotate(int updown); 
	void ZBumpRotate(int updown); 
	void SetProjectionType(int projectiontype);
	void SetEye(Vector3 neweye);
	void SetFromUp(Vector3 newfrom, Vector3 newup);
	void SetEyeFromUp(Vector3 neweye,Vector3 newfrom, Vector3 newup);
	//Accessors
	Matrix4 View(){return _view;}
	Box2 ViewRect(){return _viewrect;}
	int ProjectionType(){return _projectiontype;}
	//Useful methods. See camera?.cpp for explanation.
	void BackFaceCull(Polyhedron3 &p);
    void InternalFaceCull(Polyhedron3 &p);//for hypercubes
	BOOL SphereClipTest(const Vector3 &testcentroid, Real testradius);
	void ZClipTest(Polyhedron3 &p);
	void Project(Poly3 &p);
	void Project(Polyhedron3 &p);
	void SetRealWindow(Frame &frame);
};

#endif //CAMERA_HPP
