/***********************************************************
* Project:	CS298                                          *
* File:		vector5.hpp                                    *
* Purpose:	This file contains information about 2-tuple,  *
*			3-tuple, and 4-tuple vectors.                  *
************************************************************/

#ifndef VECTOR_HPP
#define VECTOR_HPP
#include "types.h"

//--TWO DIMENSIONAL STUFF------

class Matrix3; //So Vector2 can mention operator*(Matrix3, Vector2)

class Vector2
{
private:
	Real _x,_y;
	friend class Frame; //For Frame::RealToPixel and Frame::PixelToReal
public:
	Vector2(){_x = _y = 0.0;}
	Vector2(Real ix, Real iy){_x = ix; _y = iy;}
	void Set(Real ix, Real iy){_x = ix; _y = iy;}
	Real x()const {return _x;}
	Real y()const {return _y;}
	Vector2 operator=(const Vector2 &u){_x=u._x; _y=u._y; return *this;}
	Vector2 operator+=(const Vector2 &u){_x+=u._x; _y+=u._y; return *this;}
	Vector2 operator-=(const Vector2 &u){_x-=u._x; _y-=u._y; return *this;}
	Vector2 operator*=(Real f){_x*=f; _y*=f; return *this;}
	Vector2 operator/=(Real f);
	Real Magnitude(); //Return the Vector2's length.
	Vector2 Normalize(); //Make yourself of unit length and return self.
	friend Vector2 operator*(const Vector2 &u, Real f); //Scalar prod
	friend Vector2 operator*(Real f, const Vector2 &u); //Scalar prod
	friend Vector2 operator+(const Vector2 &u, const Vector2 &v); //Vector2 sum
	friend Vector2 operator-(const Vector2 &u, const Vector2 &v); //Vector2 difference
	friend Real operator%(const Vector2 &u, const Vector2 &v); //Dot product
	friend int operator==(const Vector2 &u, const Vector2 &v) //equality predicate.
		{if (u._x == v._x && u._y == v._y) return 1; else return 0;}
     friend Vector2 operator*(const Matrix3 &M, const Vector2 &V);
};

class Poly2; //Matrix3 needs to know there is such a class to declare
	//the operator*(Matrix, Poly2) as a friend.

class Matrix3
{
private:
	Real _m[3][3];
     /*Note that _m[i][j] means the entry in the i_th row and the j_th
     column, so this is similar to usual matrix notation like m_ij.  One
     minor confusion is that C language indices start at 0, but matrix
     indices usually start at 1.*/
public:
	Matrix3(){Identity();} //Call the Identity operation at construction.
     void Identity(); //Set to the identity matrix.
     /* The following three friend operations create special kinds
     of matrices*/
	void SetTranslationRotationScale(Vector2 trans, Real ang, Real s);
	friend Matrix3 ScaleMatrix(Real sx, Real sy);
     	//This scales x and y by factors sx and sy.
     friend Matrix3 RotationMatrix(Real ang);
     	//Rotate counterclockwise by ang radians.
     friend Matrix3 TranslationMatrix(Vector2 trans);
     	//Translates by vector trans.
     friend Matrix3 operator*(const Matrix3 &M, const Matrix3 &N);
     /* The following two friend operations are used to multiply
     in the matrix on the left, as in V = M * U for U and V vectors.*/
     friend Vector2 operator*(const Matrix3 &M, const Vector2 &V);
     friend Poly2 operator*(const Matrix3 &m, const Poly2 &v);
};

//-----THREE DIMENSIONAL STUFF-------
class Matrix4; //So Vector3 can mention operator*(Matrix4, Vector3)

class Vector3
{
private:
	Real _x,_y,_z;
	friend class Frame; //For Frame::RealToPixel and Frame::PixelToReal
	friend Matrix4;
public:
	Vector3(){_x = _y = _z = 0.0;}
	Vector3(Real ix, Real iy, Real iz){_x = ix; _y = iy; _z = iz;}
	void Set(Real ix, Real iy, Real iz){_x = ix; _y = iy; _z = iz;}
	Real x()const {return _x;}
	Real y()const {return _y;}
	Real z()const {return _z;}
	Vector3 operator=(const Vector3 &u){_x=u._x; _y=u._y; _z=u._z;
		return *this;}
	Vector3 operator+=(const Vector3 &u){_x+=u._x; _y+=u._y; _z+=u._z;
		return *this;}
	Vector3 operator-=(const Vector3 &u){_x-=u._x; _y-=u._y; _z-=u._z;
		return *this;}
	Vector3 operator*=(Real f){_x*=f; _y*=f; _z*=f; return *this;}
	Vector3 operator/=(Real f);
	Real Magnitude(); //Return the Vector3's length.
	Vector3 Normalize(); //Make yourself of unit length and return self.
	Real ReturnMagnitudeAndNormalize(); //Return old mag, and normalize.
	friend Vector3 operator*(const Vector3 &u, Real f); //Scalar prod
	friend Vector3 operator*(Real f, const Vector3 &u); //Scalar prod
	friend Vector3 operator+(const Vector3 &u, const Vector3 &v); //Vector3 sum
	friend Vector3 operator-(const Vector3 &u, const Vector3 &v); //Vector3 difference
	friend Real operator%(const Vector3 &u, const Vector3 &v); //Dot product
	//Determinant is used in solving the equation which finds the barycentric coordinates of a vertex
	//So, it is related to the Fourth dimension project. 
	friend Real Determinant(const Vector3 &A, const Vector3 &B, const Vector3 &C, const Vector3 &D);
	friend Vector3 operator*(const Vector3 &u, const Vector3 &v); //Cross product
	friend int operator==(const Vector3 &u, const Vector3 &v) //equality predicate.
	 {if (u._x == v._x && u._y == v._y && u._z == v._z) return 1;
		else return 0;}
     friend Vector3 operator*(const Matrix4 &M, const Vector3 &V);
};


class Poly3; //Matrix4 needs to know there is such a class to declare
	//the operator*(Matrix, Poly3) as a friend.
class Polyhedron3; //Matrix4 needs to know there is such a class to declare
	//the operator*(Matrix, Polyhedron3) as a friend.

class Matrix4
{
private:
	Real _m[4][4];
     /*Note that _m[i][j] means the entry in the i_th row and the j_th
     column, so this is similar to usual matrix notation like m_ij.  One
     minor confusion is that C language indices start at 0, but matrix
     indices usually start at 1.*/
public:
	Matrix4(){Identity();} //Call the Identity operation at construction.
     void Identity(); //Set to the identity matrix.
	//The default destructor frees the static _m memory.
	Matrix4 operator=(const Matrix4 &A); //Just to be sure.
	void SetTranslationRotationScale(Vector3 trans, const Matrix4 &rot,
		Real s = 1.0);
	void SetTranslationRotation(Vector3 trans, const Matrix4 &rot);
	void SetColumns(const Vector3 &u, const Vector3 &v, const Vector3 &w);
	Matrix4 Invert(); //Invert rigid-body matrix and return the inverse.
     /* The following friend operations create special matrices*/
     friend Matrix4 XRotationMatrix(Real ang);
     friend Matrix4 YRotationMatrix(Real ang);
     friend Matrix4 ZRotationMatrix(Real ang);
     friend Matrix4 TranslationMatrix(Vector3 trans);
     	//Translates by vector trans.
     friend Matrix4 operator*(const Matrix4 &M, const Matrix4 &N);
     /* The following two friend operations are used to multiply
     in the matrix on the left, as in V = M * U for U and V vectors.*/
     friend Vector3 operator*(const Matrix4 &M, const Vector3 &V);
     friend Poly3 operator*(const Matrix4 &m, const Poly3 &v);
     friend Polyhedron3 operator*(const Matrix4 &M, const Polyhedron3 &p);
};

//-----FOUR DIMENSIONAL STUFF-------
class Polytope4;

class Matrix5; //So Vector4 can mention operator*(Matrix4, Vector3)

class Vector4
{
private:
	Real _x,_y,_z,_w;
	friend class Frame; //For Frame::RealToPixel and Frame::PixelToReal
	friend Matrix5;
public:
	Vector4(){_x = _y = _z =_w = 0.0;}
	Vector4(Real ix, Real iy, Real iz, Real iw){_x = ix; _y = iy; _z = iz; _w = iw;}
	void Set(Real ix, Real iy, Real iz, Real iw){_x = ix; _y = iy; _z = iz; _w =iw;}
	Real x()const {return _x;}
	Real y()const {return _y;}
	Real z()const {return _z;}
	Real w()const {return _w;}
	Vector4 operator=(const Vector4 &u){_x=u._x; _y=u._y; _z=u._z;_w=u._w;
		return *this;}
	Vector4 operator+=(const Vector4 &u){_x+=u._x; _y+=u._y; _z+=u._z;_w+=u._w;
		return *this;}
	Vector4 operator-=(const Vector4 &u){_x-=u._x; _y-=u._y; _z-=u._z; _w-=u._w;
		return *this;}
	Vector4 operator*=(Real f){_x*=f; _y*=f; _z*=f; _w*=f; return *this;}
	Vector4 operator/=(Real f);
	Real Magnitude(); //Return the Vector4's length.
	Vector4 Normalize(); //Make yourself of unit length and return self.
	Real ReturnMagnitudeAndNormalize(); //Return old mag, and normalize.
	friend Vector4 operator*(const Vector4 &u, Real f); //Scalar prod
	friend Vector4 operator*(Real f, const Vector4 &u); //Scalar prod
	friend Vector4 operator+(const Vector4 &u, const Vector4 &v); //Vector4 sum
	friend Vector4 operator-(const Vector4 &u, const Vector4 &v); //Vector4 difference
	friend Real operator%(const Vector4 &u, const Vector4 &v); //Dot product
	friend Vector4 operator*(const Vector4 &u, const Vector4 &v); //simple product
	friend Vector4 TripleCross(const Vector4 &u, const Vector4 &v, const Vector4 &s); //Cross product
	friend int operator==(const Vector4 &u, const Vector4 &v) //equality predicate.
	 {if (u._x == v._x && u._y == v._y && u._z == v._z && u._w == v._w) return 1;
		else return 0;}
    friend Vector4 operator*(const Matrix5 &M, const Vector4 &V);
};

class Polytope4; //Matrix5 needs to know there is such a class to declare
    //the operator*(Matrix, Polytope4) as a friend.   

class Matrix5
{
private:
	Real _m[5][5];
     /*Note that _m[i][j] means the entry in the i_th row and the j_th
     column, so this is similar to usual matrix notation like m_ij.  One
     minor confusion is that C language indices start at 0, but matrix
     indices usually start at 1.*/
public:
	Matrix5(){Identity();} //Call the Identity operation at construction.
     void Identity(); //Set to the identity matrix.
	//The default destructor frees the static _m memory.
	Matrix5 operator=(const Matrix5 &A); //Just to be sure.
	void SetTranslationRotationScale(Vector4 trans, const Matrix5 &rot,
		Real s = 1.0);
	void SetTranslationRotation(Vector4 trans, const Matrix5 &rot);
	void SetColumns(const Vector4 &u, const Vector4 &v,const Vector4 &w, const Vector4 &s);
	Matrix5 Invert(); //Invert rigid-body matrix and return the inverse.
     /* The following friend operations create special matrices for rotation*/
     friend Matrix5 P12RotationMatrix5(Real ang); 
     friend Matrix5 P13RotationMatrix5(Real ang); 
     friend Matrix5 P14RotationMatrix5(Real ang); 
     friend Matrix5 P23RotationMatrix5(Real ang); 
     friend Matrix5 P24RotationMatrix5(Real ang); 
     friend Matrix5 P34RotationMatrix5(Real ang); 
	 friend Matrix5 TranslationMatrix(Vector4 trans);
     	//Translates by vector trans.
     friend Matrix5 operator*(const Matrix5 &M, const Matrix5 &N);
     /* The following two friend operations are used to multiply
     in the matrix on the left, as in V = M * U for U and V vectors.*/
     friend Vector4 operator*(const Matrix5 &M, const Vector4 &V);
     friend Polytope4 operator*(const Matrix5 &M, const Polytope4 &P);

};

#endif //VECTOR_HPP
